# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Autogenerated Cargo.toml configuration file
# This file was generated by OptimizationEngine
#
# See https://alphaville.github.io/optimization-engine/
#
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
[package]
name = "{{meta.optimizer_name}}"
version = "{{meta.version}}"
license = "{{meta.licence}}"
authors = {{meta.authors | tojson }}
edition = "2018"
publish=false


[dependencies]
{% if build_config.local_path is not none -%}
optimization_engine = {path = "{{build_config.local_path}}"{%- if build_config.allocator.value == 1%}, features=["jem"]{% elif build_config.allocator.value == 2 %}, features=["rp"]{% endif %}}
{% else -%}
optimization_engine = "{{build_config.open_version or '*'}}{%- if build_config.allocator.value == 1%}, features=["jem"]{% elif build_config.allocator.value == 2 %}, features=["rp"]{% endif %}"
{% endif %}

icasadi_{{meta.optimizer_name}} = {path = "./icasadi_{{meta.optimizer_name}}/"}
{% if activate_clib_generation -%}
libc = "0.2"

[build-dependencies]
cbindgen = "0.20"
{% endif %}


{% if activate_clib_generation -%}
[lib]
crate-type = ["staticlib", "cdylib", "rlib"]
{% else %}
[lib]
crate-type = ["rlib"]
{% endif %}
